import dataclasses
import functools
import typing

from . import models
from . import utils

@dataclasses.dataclass
class Repository:
    store: set = dataclasses.field(default_factory = set)

    def __iter__(self) -> typing.Iterable:
        return iter(self.store)

    def __repr__(self) -> str:
        return '{class_name}[{cases}]'.format \
        (
            class_name = type(self).__name__,
            cases      = ', '.join(map(repr, self))
        )

    def add(self, item: typing.Any) -> None:
        return self.store.add(item)

    def clear(self) -> None:
        return self.store.clear()

    def copy(self):
        clone = self.__class__()

        clone.update(self.store)

        return clone

    def pop(self) -> typing.Any:
        return self.store.pop()

    def remove(self, item: typing.Any) -> None:
        return self.store.remove(item)

    def update(self, *iterables: typing.Iterable) -> None:
        return self.store.update(*iterables)

class Cases(Repository):
    def __call__ \
            (
                self,
                cls_or_fn: typing.Optional[typing.Union[typing.Type, typing.Callable]] = None,
                /,
                *,
                name: typing.Optional[str] = None,
            ):
        def wrap(cls_or_fn: typing.Union[typing.Type[models.Case], typing.Callable]):
            case_style: typing.Optional[str] = \
            (
                cls_or_fn.__name__.lower()
                if name is None and utils.named(cls_or_fn)
                else name
            )

            case: models.Case

            if isinstance(cls_or_fn, type):
                @functools.wraps(cls_or_fn, updated = ())
                @models.model
                class Wrapper(models.model(cls_or_fn)):
                    style: typing.Optional[str] = case_style

                case = Wrapper()

                cls_or_fn = Wrapper
            elif callable(cls_or_fn):
                case = models.Case \
                (
                    render = cls_or_fn,
                    style  = case_style,
                )

            self.add(case)

            return cls_or_fn

        return \
        (
            wrap
            if cls_or_fn is None
            else wrap(cls_or_fn)
        )

    def get(self, style: str, default: typing.Any = None) -> typing.Any:
        items: typing.List[models.Case] = \
        [
            case
            for case in self
            if case.style == style
        ]

        return \
        (
            items[0]
            if items
            else default
        )

    def identify(self, string: str) -> typing.List[models.Case]:
        return \
        [
            case
            for case in self
            if case.match(string)
        ]



from . import models
from . import renderers
from . import repositories
from . import translators
from . import types

cases: repositories.Cases = repositories.Cases()

@cases
class Lower(models.Case):
    prepare: types.Translator = translators.lower
    render:  types.Renderer   = renderers.space

@cases
class Upper(models.Case):
    prepare: types.Translator = translators.upper
    render:  types.Renderer   = renderers.space

@cases
class Title(models.Case):
    prepare: types.Translator = translators.title
    render:  types.Renderer   = renderers.space

@cases
class Sentence(models.Case):
    prepare: types.Translator = translators.capitalize
    render:  types.Renderer   = renderers.space

@cases
class Snake(models.Case):
    prepare: types.Translator = translators.lower
    render:  types.Renderer   = renderers.underscore

@cases
class Helter(models.Case):
    prepare: types.Translator = translators.title
    render:  types.Renderer   = renderers.underscore

@cases
class Macro(models.Case):
    prepare: types.Translator = translators.upper
    render:  types.Renderer   = renderers.underscore

@cases
class Kebab(models.Case):
    prepare: types.Translator = translators.lower
    render:  types.Renderer   = renderers.hyphen

@cases
class Train(models.Case):
    prepare: types.Translator = translators.title
    render:  types.Renderer   = renderers.hyphen

@cases
class Cobol(models.Case):
    prepare: types.Translator = translators.upper
    render:  types.Renderer   = renderers.hyphen

@cases
class Flat(models.Case):
    prepare: types.Translator = translators.lower
    render:  types.Renderer   = renderers.concatenate

@cases
class Flush(models.Case):
    prepare: types.Translator = translators.upper
    render:  types.Renderer   = renderers.concatenate

@cases
class Pascal(models.Case):
    prepare: types.Translator = translators.title
    render:  types.Renderer   = renderers.concatenate

@cases
class Camel(models.Case):
    prepare: types.Translator = translators.dromedary
    render:  types.Renderer   = renderers.concatenate

@cases
class Dot(models.Case):
    prepare: types.Translator = translators.lower
    render:  types.Renderer   = renderers.period

@cases
class Path(models.Case):
    prepare: types.Translator = translators.lower
    render:  types.Renderer   = renderers.slash



from .models       import Case
from .parsers      import Parser
from .repositories import Cases
from .styles       import *

parser: Parser = Parser()

parse = parser.parse

identify = cases.identify
case     = cases.__call__

locals().update \
(
    {
        case.style: case
        for case in cases
    }
)



import random
import typing
import itertools

from . import utils
from . import renderers

def lower(words: typing.List[str]) -> typing.List[str]:
    return [word.lower() for word in words]

def upper(words: typing.List[str]) -> typing.List[str]:
    return [word.upper() for word in words]

def title(words: typing.List[str]) -> typing.List[str]:
    return [word.title() for word in words]

def swapcase(words: typing.List[str]) -> typing.List[str]:
    return [word.swapcase() for word in words]

def capitalize(words: typing.List[str]) -> typing.List[str]:
    return \
    [
        (
            word.title()
            if index == 0
            else word.lower()
        )
        for index, word in enumerate(words)
    ]

def dromedary(words: typing.List[str]) -> typing.List[str]:
    return \
    [
        (
            word.lower()
            if index == 0
            else word.title()
        )
        for index, word in enumerate(words)
    ]

def alternating(words: typing.List[str]) -> typing.List[str]:
    return utils.chunk \
    (
        string = ''.join \
        (
            (str.lower, str.upper)[index % 2](character)
            for index, character in enumerate(itertools.chain(*words))
        ),
        sizes  = [len(word) for word in words],
    )

def sponge(words: typing.List[str]) -> typing.List[str]:
    return utils.chunk \
    (
        string = renderers.concatenate \
        (
            random.choice((str.lower, str.upper))(character)
            for word      in words
            for character in word
        ),
        sizes  = [len(word) for word in words],
    )



import typing

Parser     = typing.Callable[[str], typing.Iterable[str]]
Renderer   = typing.Callable[[typing.Iterable[str]], str]
Translator = typing.Callable[[typing.Iterable[str]], typing.Iterable[str]]



from . import types

space:       types.Renderer = ' '.join
concatenate: types.Renderer = ''.join
underscore:  types.Renderer = '_'.join
hyphen:      types.Renderer = '-'.join
period:      types.Renderer = '.'.join
slash:       types.Renderer = '/'.join



import dataclasses
import typing

from . import types
from . import parsers
from . import utils

model = dataclasses.dataclass \
(
    frozen = True,
    repr   = False,
)

@model
class Case:
    render:  types.Renderer
    prepare: types.Translator     = utils.identity
    parse:   types.Parser         = dataclasses.field(default_factory = parsers.Parser)
    style:   typing.Optional[str] = None

    def __repr__(self) -> str:
        return '{class_name}(style={style!r})'.format \
        (
            class_name = type(self).mro()[-2].__name__,
            style      = self.style,
        )

    def __call__(self, string: str) -> str:
        return self.render(self.prepare(self.parse(string)))

    def match(self, string: str) -> bool:
        return self(string) == string



import abc
import dataclasses
import re
import string
import typing

class AbstractParser(abc.ABC):
    @abc.abstractmethod
    def parse(self, string: str) -> typing.List[str]:
        raise NotImplementedError

@dataclasses.dataclass(frozen = True)
class Parser(AbstractParser):
    tall:  str = string.ascii_uppercase
    small: str = string.ascii_lowercase + string.digits

    def __call__(self, string: str) -> typing.List[str]:
        return self.parse(string)

    def _split_sentence(self, sentence: str) -> typing.List[str]:
        return re.findall \
        (
            pattern = f'[{self.tall}{self.small}]+',
            string  = sentence,
        )

    def _split_word(self, word: str) -> typing.List[str]:
        words: typing.List[str] = []

        current: str
        for current in re.split(f'([{self.tall}][{self.small}]*)', word):
            if not current: continue

            if not words or not all(item[-1].isupper() for item in (current, words[-1]) if item):
                words.append(str())

            words[-1] += current

        return words

    def parse(self, string: str, case_sensitive: bool = False) -> typing.List[str]:
        return \
        [
            (
                word
                if case_sensitive
                else word.lower()
            )
            for segment in self._split_sentence(string)
            for word in self._split_word(segment)
        ]



import io
import typing
import types

def compose(*functions: typing.Callable) -> typing.Callable:
    def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        response: typing.Any = None

        index: int
        function: typing.Callable
        for index, function in enumerate(functions):
            response = \
            (
                function(*args, **kwargs)
                if not index
                else function(response)
            )

        return response

    return wrapper

def identity(x: typing.Any) -> typing.Any:
    return x

def is_lambda(obj: typing.Any) -> bool:
    return isinstance(obj, types.LambdaType) and obj.__name__ == '<lambda>'

def chunk(string: str, sizes: typing.List[int]) -> typing.List[str]:
    buffer: io.StringIO = io.StringIO(string)

    return \
    [
        buffer.read(size)
        for size in sizes
    ]

def named(obj: typing.Any) -> bool:
    return hasattr(obj, '__name__')
